"""proxy_test.py"""
# pylint: disable=missing-docstring,line-too-long

from nose.tools import assert_raises # pylint: disable=no-name-in-module

from .proxy import Proxy
from . import exceptions

def test_ConnectionError():
    proxy = Proxy('http://non-existant.com.au/')
    with assert_raises(exceptions.ConnectionError):
        proxy.add(1, 2)



"""exceptions.py"""

class RPCClientException(Exception):
    """Base class for the other rpcclient exceptions"""
    pass

class ConnectionError(RPCClientException): # pylint: disable=redefined-builtin
    """The requests module raised an error, which could be a network issue,
    invalid HTTP response or timeout. See
    http://stackoverflow.com/questions/16511337/"""

    def __init__(self):
        super().__init__('Connection error')

class StatusCodeError(RPCClientException):
    """The server responded with status code != 200"""

    def __init__(self, status_code):
        super().__init__('Returned status code '+str(status_code))

class ParseError(RPCClientException):
    """Couldnt parse the json response (invalid json)"""

    def __init__(self):
        super().__init__('Parse error')

class InvalidResponse(RPCClientException):
    """The response didnt validate against the JSON-RPC response schema"""

    def __init__(self):
        super().__init__('Invalid response')

class ReceivedNoResponse(RPCClientException):
    """A response was expected, but no response was received"""

    def __init__(self):
        super().__init__('No response')

class ReceivedErrorResponse(RPCClientException):
    """The server responded with 'error'. Raise it so it must be handled"""

    def __init__(self, code, message): #pylint:disable=unused-argument
        super().__init__(message)



"""exceptions_test.py"""
# pylint: disable=missing-docstring,line-too-long

from nose.tools import assert_raises # pylint: disable=no-name-in-module

from . import exceptions

def test_ConnectionError():
    with assert_raises(exceptions.ConnectionError):
        raise exceptions.ConnectionError

def test_StatusCodeError():
    with assert_raises(exceptions.StatusCodeError):
        raise exceptions.StatusCodeError(404)

def test_ParseError():
    with assert_raises(exceptions.ParseError):
        raise exceptions.ParseError

def test_InvalidResponse():
    with assert_raises(exceptions.InvalidResponse):
        raise exceptions.InvalidResponse

def test_ReceivedNoResponse():
    with assert_raises(exceptions.ReceivedNoResponse):
        raise exceptions.ReceivedNoResponse

def test_ReceivedErrorResponse():
    with assert_raises(exceptions.ReceivedErrorResponse):
        raise exceptions.ReceivedErrorResponse(1, 'Member not found')



"""rpc_test.py

This should test:

Notifications
    rpc.request('go')
    rpc.request('print', 'Hello world')
    rpc.request('print', 'Hello', 'world')
    rpc.request('print', string='Hello world')
    rpc.request('print', string1='Hello', string2='world')
    rpc.request('set_age', 40, name='Beau Barker')

Requests (requiring a response):
    rpc.request('get', response=True)
    rpc.request('sqrt', 1, response=True)
    rpc.request('add', 1, 2, response=True)
    rpc.request('find', name='Beau', response=True)
    rpc.request('find', name='Beau', age=38, response=True)
    rpc.request('set_age', 40, name="Beau Barker", response=True)
"""
# pylint: disable=missing-docstring,line-too-long

import json
from nose.tools import assert_equal # pylint: disable=no-name-in-module

from . import rpc

# Notifications

def test_method_only():

    assert_equal(
        '{"jsonrpc": "2.0", "method": "go"}',
        json.dumps(rpc.request('go'))
    )

def test_one_arg():

    assert_equal(
        '{"jsonrpc": "2.0", "method": "print", "params": ["Hello world"]}',
        json.dumps(rpc.request('print', 'Hello world'))
    )

def test_two_args():

    assert_equal(
        '{"jsonrpc": "2.0", "method": "print", "params": ["Hello", "world"]}',
        json.dumps(rpc.request('print', 'Hello', 'world'))
    )

def test_one_kwarg():

    assert_equal(
        '{"jsonrpc": "2.0", "method": "print", "params": {"string": "Hello world"}}',
        json.dumps(rpc.request('print', string='Hello world'))
    )

def test_two_kwargs():

    assert_equal(
        '{"jsonrpc": "2.0", "method": "print", "params": {"string1": "Hello", "string2": "world"}}',
        json.dumps(rpc.request('print', string1='Hello', string2='world'))
    )

def test_args_and_kwargs():
    # Both args and kwargs

    assert_equal(
        '{"jsonrpc": "2.0", "method": "set_age", "params": [40, {"name": "Beau Barker"}]}',
        json.dumps(rpc.request('set_age', 40, name='Beau Barker'))
    )

def test_multiple_args_and_kwargs():
    # Both args and kwargs

    assert_equal(
        '{"jsonrpc": "2.0", "method": "set_age", "params": [39, 40, {"first": "Beau", "last": "Barker"}]}',
        json.dumps(rpc.request('set_age', 39, 40, first='Beau', last='Barker'))
    )

# Non-notifications (require a response)

def test_method_only_request():

    assert_equal(
        '{"jsonrpc": "2.0", "method": "get", "id": 1}',
        json.dumps(rpc.request('get', response=True))
    )

def test_one_arg_request():

    assert_equal(
        '{"jsonrpc": "2.0", "method": "sqrt", "params": [1], "id": 2}',
        json.dumps(rpc.request('sqrt', 1, response=True))
    )

def test_two_args_request():

    assert_equal(
        '{"jsonrpc": "2.0", "method": "add", "params": [1, 2], "id": 3}',
        json.dumps(rpc.request('add', 1, 2, response=True))
    )

def test_one_kwarg_request():

    assert_equal(
        '{"jsonrpc": "2.0", "method": "find", "params": {"name": "Beau"}, "id": 4}',
        json.dumps(rpc.request('find', name='Beau', response=True))
    )

def test_two_kwargs_request():
    # Note in the resulting json, the params are in reverse order to how the
    # kwargs are called, because the args are sorted by key in the request
    # function. This is the only way to know exactly how the request will look
    # because the kwargs come in a random order. There is no way of getting the
    # actual order of the kwargs

    assert_equal(
        '{"jsonrpc": "2.0", "method": "find", "params": {"age": 38, "name": "Beau"}, "id": 5}',
        json.dumps(rpc.request('find', name='Beau', age=38, response=True))
    )

def test_args_and_kwargs_request():
    # Includes both args and kwargs

    assert_equal(
        '{"jsonrpc": "2.0", "method": "find", "params": {"age": 38, "name": "Beau"}, "id": 6}',
        json.dumps(rpc.request('find', name='Beau', age=38, response=True))
    )

def test_multiple_args_and_kwargs_request():
    # Both args and kwargs

    assert_equal(
        '{"jsonrpc": "2.0", "method": "set_age", "params": [39, 40, {"first": "Beau", "last": "Barker"}], "id": 7}',
        json.dumps(rpc.request('set_age', 39, 40, first='Beau', last='Barker', response=True))
    )



"""proxy.py"""

import os
import json
import logging

import requests
import jsonschema

from . import rpc
from . import exceptions

class Proxy:
    """This class acts as the remote server"""

    def __init__(self, url):
        self.url = url

    def __getattr__(self, name):
        """Catch undefined methods and handle them as RPC requests.

        The technique is here:
        http://code.activestate.com/recipes/307618/
        """

        def handlerMethod(*args, **kwargs):
            """Send the RPC request to the proxy server.

            Calls a procedure on another server.
            Raises RPCClientException: On any error caught.
            """

            # Construct an rpc request from the method name and other arguments
            # passed
            request_dict = rpc.request(name, *args, **kwargs)

            # Log the request
            logging.info('--> '+json.dumps(request_dict))

            try:
                # Send the message
                r = requests.post(
                    self.url,
                    headers={
                        'Content-Type': 'application/json; charset=utf-8'
                    },
                    json=request_dict
                )

                logging.info('<-- '+r.text.strip("\n"))

            except (requests.exceptions.InvalidSchema,
                    requests.exceptions.RequestException):
                raise exceptions.ConnectionError()

            # Raise exception if any HTTP response other than 200
#            if r.status_code != 200:
#                raise exceptions.StatusCodeError(r.status_code)

            # A response was expected, but none was given?
            if 'id' in request_dict and not len(r.text):
                raise exceptions.ReceivedNoResponse()

            # Was response given?
            if len(r.text):

                # Attempt to parse the response
                try:
                    response_dict = json.loads(r.text)

                except ValueError:
                    raise exceptions.ParseError()

                # A response was *not* expected, but one was given? It may not
                # be necessary to raise here. If we receive a response anyway,
                # can't we just ignore it?
                if 'id' not in request_dict and 'result' in response_dict:
                    raise exceptions.InvalidResponse()

                # Validate the response against the Response schema
                try:
                    jsonschema.validate(
                        response_dict,
                        json.loads(open(os.path.dirname(__file__)+ \
                            '/response-schema.json').read()))

                except jsonschema.ValidationError:
                    raise exceptions.InvalidResponse()

                # If the response was "error", raise it, to ensure it's handled
                if 'error' in response_dict:
                    raise exceptions.ReceivedErrorResponse(
                        response_dict['error']['code'],
                        response_dict['error']['message'])

                # Otherwise, surely we have a result to return
                print(response_dict['result'])

            return None

        return handlerMethod



"""rpc.py"""

import itertools
from collections import OrderedDict

id_generator = itertools.count(1) # First generated is 1

def request(method, *args, **kwargs):
    """Returns an rpc notification, (a request not expecting a response.

    Notification examples:
        rpc.notification('go')
        rpc.notification('print', 'Hello world')
        rpc.notification('print', str='Hello world')

    Request examples (requiring a response):
        rpc.request('get', response=True)
        rpc.request('sqrt', 1, response=True)
        rpc.request('add', 1, 2, response=True)
        rpc.request('find', name='Beau', response=True)
        rpc.request('find', name='Beau', age=38, response=True)
    """

    # Get the request id, if present in kwargs, and pop it out of the dict
    request_id = None
    if 'response' in kwargs:
        kwargs.pop('response')
        request_id = next(id_generator)

    r = OrderedDict([
        ('jsonrpc', '2.0'),
        ('method', method)
    ])

    # Any params passed?
    if args or kwargs:

        params = None

        # Add any params passed as args
        if args:
            params = list(args)

        # Add any params passed as kwargs
        if kwargs:
            if params:
                params.append(OrderedDict(sorted(kwargs.items())))
            else:
                params = OrderedDict(sorted(kwargs.items()))

        r.update(OrderedDict([
            ('params', params)
        ]))

    # Add id if response=True was passed
    if request_id:
        r.update(OrderedDict([
            ('id', request_id)
        ]))

    return r



#!/usr/bin/env python
"""setup.py"""

from distutils.core import setup

setup(
    name = 'jsonrpcclient',
    packages = ['jsonrpcclient'],
    package_data={'jsonrpcclient': ['response-schema.json']},
    install_requires = ['jsonschema', 'requests'],
    version = '0.01',
    description = 'JSON-RPC 2.0 client library',
    author = 'Beau Barker',
    author_email = 'beauinmelbourne@gmail.com',
    url = 'https://bitbucket.org/beau-barker/jsonrpcclient',
    download_url = 'https://bitbucket.org/beau-barker/jsonrpcclient/get/0.01.zip',
    keywords = ['json-rpc', 'json', 'api'],
    classifiers = [
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Development Status :: 4 - Beta',
        'Environment :: Other Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules',
        ],
    )
